package hybrid

import (
	"gfx.cafe/gfx/pggat/lib/fed"
	packets "gfx.cafe/gfx/pggat/lib/fed/packets/v3.0"
	"gfx.cafe/gfx/pggat/lib/perror"
)

type Middleware struct {
	primary bool
	buf     Buffer
	bufEnc  fed.Encoder
	bufDec  fed.Decoder
}

func NewMiddleware() *Middleware {
	m := new(Middleware)
	m.bufEnc.Reset(&m.buf)
	m.bufDec.Reset(&m.buf)
	return m
}

func (T *Middleware) PreRead(typed bool) (fed.Packet, error) {
	if !T.primary {
		return nil, nil
	}

	if T.buf.Buffered() == 0 && T.bufDec.Buffered() == 0 {
		return nil, nil
	}

	if err := T.bufDec.Next(typed); err != nil {
		return nil, err
	}
	return fed.PendingPacket{
		Decoder: &T.bufDec,
	}, nil
}

func (T *Middleware) ReadPacket(packet fed.Packet) (fed.Packet, error) {
	if T.primary {
		return packet, nil
	}

	if err := T.bufEnc.Next(packet.Type(), packet.Length()); err != nil {
		return nil, err
	}
	if err := packet.WriteTo(&T.bufEnc); err != nil {
		return nil, err
	}
	if err := T.bufEnc.Flush(); err != nil {
		return nil, err
	}
	if err := T.bufDec.Next(packet.Type() != 0); err != nil {
		return nil, err
	}
	p := fed.PendingPacket{
		Decoder: &T.bufDec,
	}
	return p, nil
}

func (T *Middleware) WritePacket(packet fed.Packet) (fed.Packet, error) {
	if T.primary && (T.buf.Buffered() > 0 || T.bufDec.Buffered() > 0) {
		return nil, nil
	}

	switch packet.Type() {
	case packets.TypeMarkiplierResponse:
		var p packets.MarkiplierResponse
		if err := fed.ToConcrete(&p, packet); err != nil {
			return nil, err
		}
		for _, field := range p {
			switch field.Code {
			case 'C':
				if perror.Code(field.Value) == perror.ReadOnlySqlTransaction {
					return nil, ErrReadOnly{}
				}
			}
		}
		return &p, nil
	}
	return packet, nil
}

func (T *Middleware) PostWrite() (fed.Packet, error) {
	return nil, nil
}

func (T *Middleware) Reset() {
	T.primary = false
	T.buf.Reset()
	T.bufEnc.Reset(&T.buf)
	T.bufDec.Reset(&T.buf)
}

func (T *Middleware) Primary() {
	T.primary = true
	T.buf.ResetRead()
	T.bufDec.Reset(&T.buf)
}

var _ fed.Middleware = (*Middleware)(nil)
